"""Instruction Finetuning API"""
from __future__ import annotations

from concurrent.futures import Future
from typing import Dict, List, Optional, Union, overload

from typing_extensions import Literal

from mcli.api.engine.engine import get_return_response, run_singular_mapi_request
from mcli.api.runs import Run
from mcli.models.finetune_config import FinetuneConfig

QUERY_FUNCTION = 'createFinetune'
VARIABLE_DATA_NAME = 'createFinetuneData'
# This returns the same data that the create_run function returns
# for consistency when rendering the describe output
QUERY = f"""
mutation CreateFinetune(${VARIABLE_DATA_NAME}: CreateFinetuneInput!) {{
  {QUERY_FUNCTION}({VARIABLE_DATA_NAME}: ${VARIABLE_DATA_NAME}) {{
    id
    name
    createdByEmail
    status
    createdAt
    startedAt
    completedAt
    updatedAt
    reason
    priority
    maxRetries
    preemptible
    retryOnSystemFailure
    resumptions {{
        clusterName
        cpus
        gpuType
        gpus
        nodes
        executionIndex
        startTime
        endTime
        status
    }}
    details {{
        originalRunInput
        metadata
        lastExecutionId
    }}
  }}
}}"""


@overload
def instruction_finetune(model: str,
                         train_data_path: str,
                         save_folder: str,
                         cluster: str,
                         eval_data_path: Optional[str] = None,
                         training_duration: Optional[str] = None,
                         learning_rate: Optional[float] = None,
                         context_length: Optional[int] = None,
                         experiment_trackers: Optional[List[Dict]] = None,
                         gpu_type: Optional[str] = None,
                         gpus: Optional[int] = None,
                         *,
                         timeout: Optional[float] = 10,
                         future: Literal[False] = False) -> Run:
    ...


@overload
def instruction_finetune(model: str,
                         train_data_path: str,
                         save_folder: str,
                         cluster: str,
                         eval_data_path: Optional[str] = None,
                         training_duration: Optional[str] = None,
                         learning_rate: Optional[float] = None,
                         context_length: Optional[int] = None,
                         experiment_trackers: Optional[List[Dict]] = None,
                         gpu_type: Optional[str] = None,
                         gpus: Optional[int] = None,
                         *,
                         timeout: Optional[float] = 10,
                         future: Literal[False] = False) -> Future[Run]:
    ...


def instruction_finetune(model: str,
                         train_data_path: str,
                         save_folder: str,
                         cluster: str,
                         eval_data_path: Optional[str] = None,
                         training_duration: Optional[str] = None,
                         learning_rate: Optional[float] = None,
                         context_length: Optional[int] = None,
                         experiment_trackers: Optional[List[Dict]] = None,
                         gpu_type: Optional[str] = None,
                         gpus: Optional[int] = None,
                         *,
                         timeout: Optional[float] = 10,
                         future: Literal[False] = False) -> Union[Run, Future[Run]]:
    """Finetunes a model on an instruction finetuning dataset and converts the final composer checkpoint to a
    Hugging Face formatted checkpoint for inference.

    Args:
        model: The name of the Hugging Face model to use. Current available options are 
            ``mosaicml/mpt-7b-8k``, ``mosaicml/mpt-30b``, and ``mosaicml/mpt-7b``.
        train_data_path: The name of a Hugging Face dataset (e.g. ``mosaicml/dolly_hhrlhf``) or the
            remote location of your training data (e.g. ``s3://my-bucket/my-data``). This should be a folder
            containing a ``train.jsonl`` file. The training data should be formatted with each row
            containing a ``prompt`` and ``response`` field.
        save_folder: The remote location to save the finetuned checkpoints. For example,
            if your ``save_folder`` is ``s3://my-bucket/my-checkpoints``, the finetuned Composer checkpoints
            will be saved to ``s3://my-bucket/my-checkpoints/<run-name>/checkpoints``, and Hugging Face formatted
            checkpoints will be saved to ``s3://my-bucket/my-checkpoints/<run-name>/hf_checkpoints``. The supported
            cloud provider prefixes are ``s3://``, ``gs://``, and ``oci://``.
        cluster: The name of the cluster you would like to use.
        eval_data_path: The name of a Hugging Face dataset (e.g. ``mosaicml/dolly_hhrlhf``) or the
            remote location of your evaluation data (e.g. ``s3://my-bucket/my-data``). This should be a folder
            containing a ``test.jsonl`` file. The evaluation data should be formatted with each row
            containing a ``prompt`` and ``response`` field. Default is ``None``.
        training_duration: The total duration of your finetuning run. This can be specified in batches (e.g. ``100ba``),
            epochs (e.g. ``10ep``), or tokens (e.g. ``1_000_000tok``). Default is ``1ep``.
        learning_rate: The peak learning rate to use for finetuning. Default is ``5e-7``. The optimizer used
            is DecoupledLionW with betas of 0.99 and 0.95 and no weight decay, and the learning rate scheduler used is
            LinearWithWarmupSchedule with a warmup of 2% of the total training duration and a final learning rate 
            multiplier of 0.
        context_length: The maximum sequence length to use. This will be applied to the model configuration and be 
            used to truncate any data that is too long. The default is the default for the provided Hugging Face model.
            The MPT model series support changing the context length from the original pretraining context 
            length through the use of ALiBi.
        experiment_trackers: A list of experiment tracker configurations. For example, to add Weights and Biases
            tracking, you can pass in ``{'integration_type': 'wandb', 'project': 'my-project', 'entity': 'my-entity'}``.
        gpu_type: The type of GPU to use. Default is the GPU type of the provided cluster.
        gpus: The number of GPUs to use. Default is a single node of the provided cluster.
        timeout: Time, in seconds, in which the call should complete. If the run creation
            takes too long, a TimeoutError will be raised. If ``future`` is ``True``, this
            value will be ignored.
        future: Return the output as a :class:`~concurrent.futures.Future`. If True, the
            call to `create_run` will return immediately and the request will be
            processed in the background. This takes precedence over the ``timeout``
            argument. To get the :type Run: output, use ``return_value.result()``
            with an optional ``timeout`` argument.

    Returns:
        A Run object containing the run information.
    """
    config = FinetuneConfig.from_dict({
        'model': model,
        'train_data_path': train_data_path,
        'save_folder': save_folder,
        'cluster': cluster,
        'eval_data_path': eval_data_path,
        'training_duration': training_duration,
        'experiment_trackers': experiment_trackers,
        'gpu_type': gpu_type,
        'gpus': gpus,
        'learning_rate': learning_rate,
        'context_length': context_length,
    })
    finetune_config = config.to_create_finetune_api_input()
    variables = {
        VARIABLE_DATA_NAME: finetune_config,
    }

    response = run_singular_mapi_request(
        query=QUERY,
        query_function=QUERY_FUNCTION,
        return_model_type=Run,
        variables=variables,
    )
    return get_return_response(response, future=future, timeout=timeout)
